@@ -26,12 +26,12 @@ __copyright__ = "Copyright (c) 2005-2009 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import pyatspi
-from gi.repository import Atspi, Atk
 
 from . import braille
 from . import debug
 from . import generator
 from . import messages
+from . import object_properties
 from . import orca_state
 from . import settings
 from . import settings_manager
@@ -56,6 +56,8 @@ class BrailleGenerator(generator.Generator):
     SKIP_CONTEXT_ROLES = (pyatspi.ROLE_MENU,
                           pyatspi.ROLE_MENU_BAR,
                           pyatspi.ROLE_PAGE_TAB_LIST,
+                          pyatspi.ROLE_REDUNDANT_OBJECT,
+                          pyatspi.ROLE_UNKNOWN,
                           pyatspi.ROLE_COMBO_BOX)
 
     def __init__(self, script):
@@ -130,11 +132,17 @@ class BrailleGenerator(generator.Generator):
         empty array.  Note that a 'role' attribute in args will
         override the accessible role of the obj.
         """
+
+        if args.get('isProgressBarUpdate') \
+           and not _settingsManager.getSetting('brailleProgressBarUpdates'):
+            return []
+
         result = []
         role = args.get('role', obj.getRole())
         verbosityLevel = _settingsManager.getSetting('brailleVerbosityLevel')
 
         doNotPresent = [pyatspi.ROLE_UNKNOWN,
+                        pyatspi.ROLE_REDUNDANT_OBJECT,
                         pyatspi.ROLE_FILLER,
                         pyatspi.ROLE_EXTENDED,
                         pyatspi.ROLE_LINK]
@@ -146,41 +154,31 @@ class BrailleGenerator(generator.Generator):
         if verbosityLevel == settings.VERBOSITY_LEVEL_BRIEF:
             doNotPresent.extend([pyatspi.ROLE_ICON, pyatspi.ROLE_CANVAS])
 
-        if verbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
+        if role == pyatspi.ROLE_HEADING:
+            level = self._script.utilities.headingLevel(obj)
+            result.append(object_properties.ROLE_HEADING_LEVEL_BRAILLE % level)
+
+        elif verbosityLevel == settings.VERBOSITY_LEVEL_VERBOSE \
            and not args.get('readingRow', False) and role not in doNotPresent:
-            result.append(self.getLocalizedRoleName(obj, role))
+            result.append(self.getLocalizedRoleName(obj, **args))
         return result
 
-    @staticmethod
-    def getLocalizedRoleName(obj, role=None):
+    def getLocalizedRoleName(self, obj, **args):
         """Returns the localized name of the given Accessible object; the name
         is suitable to be brailled.
 
         Arguments:
         - obj: an Accessible object
-        - role: an optional pyatspi role to use instead
         """
 
         if _settingsManager.getSetting('brailleRolenameStyle') \
                 == settings.BRAILLE_ROLENAME_STYLE_SHORT:
-            objRole = role or obj.getRole()
-            rv = shortRoleNames.get(objRole)
+            role = args.get('role', obj.getRole())
+            rv = shortRoleNames.get(role)
             if rv:
                 return rv
 
-        if not isinstance(role, pyatspi.Role):
-            try:
-                return obj.getLocalizedRoleName()
-            except:
-                return ''
-
-        if not role:
-            return ''
-
-        nonlocalized = Atspi.role_get_name(role)
-        atkRole = Atk.role_for_name(nonlocalized)
-
-        return Atk.role_get_localized_name(atkRole)
+        return super().getLocalizedRoleName(obj, **args)
 
     def _generateUnrelatedLabels(self, obj, **args):
         result = []
@@ -202,6 +200,11 @@ class BrailleGenerator(generator.Generator):
         specifications) that represent the accelerator for the object,
         or an empty array if no accelerator can be found.
         """
+
+        verbosityLevel = _settingsManager.getSetting('brailleVerbosityLevel')
+        if verbosityLevel == settings.VERBOSITY_LEVEL_BRIEF:
+            return []
+
         result = []
         [mnemonic, shortcut, accelerator] = \
             self._script.utilities.mnemonicShortcutAccelerator(obj)
@@ -297,6 +300,9 @@ class BrailleGenerator(generator.Generator):
             if result and parentResult:
                 result.append(braille.Region(" "))
             result.extend(parentResult)
+            if role == pyatspi.ROLE_EMBEDDED:
+                break
+
             parent = parent.parent
         result.reverse()
         return result
@@ -307,23 +313,74 @@ class BrailleGenerator(generator.Generator):
         if role not in [pyatspi.ROLE_LIST, pyatspi.ROLE_LIST_BOX]:
             return result
 
-        try:
-            s = obj.querySelection()
-        except NotImplementedError:
-            isFocused = \
-                lambda x: x and x.getState().contains(pyatspi.STATE_FOCUSED)
-            items = pyatspi.utils.findAllDescendants(obj, isFocused)
+        if 'Selection' in pyatspi.listInterfaces(obj):
+            items = self._script.utilities.selectedChildren(obj)
         else:
-            items = [s.getSelectedChild(i) for i in range(s.nSelectedChildren)]
-            if not items and obj.childCount:
-                items.append(obj[0])
+            items = [self._script.utilities.focusedChild(obj)]
+        if not (items and items[0]):
+            return result
 
-        items = list(map(self._generateName, items))
-        for item in items:
+        for item in map(self._generateName, items):
             result.extend(item)
 
         return result
 
+    def _generateListBoxItemWidgets(self, obj, **args):
+        widgetRoles = [pyatspi.ROLE_CHECK_BOX,
+                       pyatspi.ROLE_COMBO_BOX,
+                       pyatspi.ROLE_PUSH_BUTTON,
+                       pyatspi.ROLE_RADIO_BUTTON,
+                       pyatspi.ROLE_SLIDER,
+                       pyatspi.ROLE_TOGGLE_BUTTON]
+        isWidget = lambda x: x and x.getRole() in widgetRoles
+        result = []
+        if obj.parent and obj.parent.getRole() == pyatspi.ROLE_LIST_BOX:
+            widgets = pyatspi.findAllDescendants(obj, isWidget)
+            for widget in widgets:
+                result.extend(self.generate(widget, includeContext=False))
+                result.append(braille.Region(" "))
+
+        return result
+
+    def _generateProgressBarIndex(self, obj, **args):
+        if not args.get('isProgressBarUpdate') \
+           or not self._shouldPresentProgressBarUpdate(obj, **args):
+            return []
+
+        acc, updateTime, updateValue = self._getMostRecentProgressBarUpdate()
+        if acc != obj:
+            number, count = self.getProgressBarNumberAndCount(obj)
+            return ['%s' % number]
+
+        return []
+
+    def _generateProgressBarValue(self, obj, **args):
+        if args.get('isProgressBarUpdate') \
+           and not self._shouldPresentProgressBarUpdate(obj, **args):
+            return []
+
+        return self._generatePercentage(obj, **args)
+
+    def _generatePercentage(self, obj, **args):
+        percent = self._script.utilities.getValueAsPercent(obj)
+        if percent is not None:
+            return ['%s%%' % percent]
+
+        return []
+
+    def _getProgressBarUpdateInterval(self):
+        interval = _settingsManager.getSetting('progressBarBrailleInterval')
+        if interval is None:
+            return super()._getProgressBarUpdateInterval()
+
+        return int(interval)
+
+    def _shouldPresentProgressBarUpdate(self, obj, **args):
+        if not _settingsManager.getSetting('brailleProgressBarUpdates'):
+            return False
+
+        return super()._shouldPresentProgressBarUpdate(obj, **args)
+
     #####################################################################
     #                                                                   #
     # Unfortunate hacks.                                                #
@@ -379,6 +436,10 @@ class BrailleGenerator(generator.Generator):
         """Returns True or False to indicate whether context should be
         included or not.
         """
+
+        if args.get('isProgressBarUpdate'):
+            return False
+
         # For multiline text areas, we only show the context if we
         # are on the very first line.  Otherwise, we show only the
         # line.
