@@ -49,6 +49,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if not self._script.utilities.inDocumentContent(obj):
             return super()._generateAncestors(obj, **args)
 
+        if self._script.inSayAll() and obj == orca_state.locusOfFocus:
+            return []
+
         result = []
         priorObj = args.get('priorObj')
         if priorObj and self._script.utilities.inDocumentContent(priorObj):
@@ -116,6 +119,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         if self._script.utilities.isZombie(obj):
             return []
 
+        if self._script.utilities.preferDescriptionOverName(obj):
+            return []
+
         role = args.get('role', obj.getRole())
         if obj != orca_state.locusOfFocus:
             if role in [pyatspi.ROLE_ALERT, pyatspi.ROLE_DIALOG]:
@@ -169,14 +175,25 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
            and not self._script.utilities.isDPub(obj):
             return []
 
+        if obj.parent and obj.name and obj.name == obj.parent.name \
+           and obj != orca_state.locusOfFocus:
+            return []
+
         # TODO - JD: Once the formatting strings are vastly cleaned up
         # or simply removed, hacks like this won't be needed.
         role = args.get('role', obj.getRole())
         if role in [pyatspi.ROLE_COMBO_BOX, pyatspi.ROLE_SPIN_BUTTON]:
             return super()._generateName(obj, **args)
 
+        if self._script.utilities.isLink(obj) \
+           and not self._script.utilities.hasExplicitName(obj):
+            return []
+
         if self._script.utilities.inDocumentContent(obj) and obj.name:
-            result = [obj.name]
+            if self._script.utilities.preferDescriptionOverName(obj):
+                result = [obj.description]
+            else:
+                result = [obj.name]
             result.extend(self.voice(speech_generator.DEFAULT))
             return result
 
@@ -244,6 +261,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
 
     # TODO - JD: Yet another dumb generator method we should kill.
     def _generateTextRole(self, obj, **args):
+        if self._script.inSayAll():
+            return []
+
         return self._generateRoleName(obj, **args)
 
     def getLocalizedRoleName(self, obj, **args):
@@ -283,6 +303,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
         force = args.get('force', False)
         start = args.get('startOffset')
         end = args.get('endOffset')
+        index = args.get('index', 0)
+        total = args.get('total', 1)
 
         if not force:
             doNotSpeak = [pyatspi.ROLE_FOOTER,
@@ -290,6 +312,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                           pyatspi.ROLE_MENU_ITEM,
                           pyatspi.ROLE_PARAGRAPH,
                           pyatspi.ROLE_SECTION,
+                          pyatspi.ROLE_REDUNDANT_OBJECT,
                           pyatspi.ROLE_UNKNOWN]
         else:
             doNotSpeak = [pyatspi.ROLE_UNKNOWN]
@@ -307,6 +330,8 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 doNotSpeak.append(pyatspi.ROLE_ALERT)
             if self._script.utilities.isAnchor(obj):
                 doNotSpeak.append(obj.getRole())
+            if total > 1:
+                doNotSpeak.append(pyatspi.ROLE_ROW_HEADER)
 
         if obj.getState().contains(pyatspi.STATE_EDITABLE):
             lastKey, mods = self._script.utilities.lastKeyAndModifiers()
@@ -316,7 +341,7 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 text = self._script.utilities.queryNonEmptyText(obj)
                 if text and end not in [None, text.characterCount]:
                     return []
-            if role in [pyatspi.ROLE_ENTRY, pyatspi.ROLE_PASSWORD_TEXT]:
+            if role in [pyatspi.ROLE_ENTRY, pyatspi.ROLE_PASSWORD_TEXT, pyatspi.ROLE_SPIN_BUTTON]:
                 result.append(self.getLocalizedRoleName(obj, **args))
             elif obj.parent and not obj.parent.getState().contains(pyatspi.STATE_EDITABLE):
                 if lastKey not in ["Home", "End", "Up", "Down", "Left", "Right", "Page_Up", "Page_Down"]:
@@ -327,15 +352,16 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 result.extend(acss)
 
         elif role == pyatspi.ROLE_HEADING:
-            level = self._script.utilities.headingLevel(obj)
-            if level:
-                result.append(object_properties.ROLE_HEADING_LEVEL_SPEECH % {
-                    'role': self.getLocalizedRoleName(obj, **args),
-                    'level': level})
-                result.extend(acss)
-            else:
-                result.append(self.getLocalizedRoleName(obj, **args))
-                result.extend(acss)
+            if index == total - 1 or not self._script.utilities.isFocusableWithMathChild(obj):
+                level = self._script.utilities.headingLevel(obj)
+                if level:
+                    result.append(object_properties.ROLE_HEADING_LEVEL_SPEECH % {
+                        'role': self.getLocalizedRoleName(obj, **args),
+                        'level': level})
+                    result.extend(acss)
+                else:
+                    result.append(self.getLocalizedRoleName(obj, **args))
+                    result.extend(acss)
 
         elif self._script.utilities.isLink(obj):
             if obj.parent.getRole() == pyatspi.ROLE_IMAGE:
@@ -345,15 +371,17 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
                 if self._script.utilities.hasUselessCanvasDescendant(obj):
                     result.append(self.getLocalizedRoleName(obj, role=pyatspi.ROLE_IMAGE))
                     result.extend(acss)
-                result.append(self.getLocalizedRoleName(obj, **args))
-                result.extend(acss)
+                if index == total - 1 or not self._script.utilities.isFocusableWithMathChild(obj):
+                    result.append(self.getLocalizedRoleName(obj, **args))
+                    result.extend(acss)
 
         elif role not in doNotSpeak and args.get('priorObj') != obj:
             result.append(self.getLocalizedRoleName(obj, **args))
             result.extend(acss)
 
-        index = args.get('index', 0)
-        total = args.get('total', 1)
+        if self._script.utilities.isMath(obj) and not self._script.utilities.isMathTopLevel(obj):
+            return result
+
         ancestorRoles = [pyatspi.ROLE_HEADING, pyatspi.ROLE_LINK]
         if index == total - 1 \
            and (role == pyatspi.ROLE_IMAGE or self._script.utilities.queryNonEmptyText(obj)):
@@ -533,6 +561,9 @@ class SpeechGenerator(speech_generator.SpeechGenerator):
             utterance = self.generateSpeech(
                 obj, startOffset=start, endOffset=end, string=string,
                 index=i, total=len(contents), **args)
+            if isinstance(utterance, list):
+                isNotEmptyList = lambda x: not (isinstance(x, list) and not x)
+                utterance = list(filter(isNotEmptyList, utterance))
             if utterance and utterance[0]:
                 result.append(utterance)
                 args['priorObj'] = obj
